#! /usr/bin/python3
# -*- coding: utf-8 -*-
"""Global default hyperparams"""

CNEIGHBOR: int = 1
SENSE_COEFFICIENT: float = 0.1
ALLOW_REIFY: bool = False
SEPARATE_1_AND_2: bool = False
USE_ALIGNMENT: bool = False
USE_SMATCH_TOP: bool = True
HANDLE_QUOTED_REENTRANCY = False
