#! /usr/bin/python3
# -*- coding: utf-8 -*-
# Author: chunjy92
# Date: 12/4/24 17:41
from __future__ import annotations

import logging

logger = logging.getLogger(__name__)


def main():
  pass


if __name__ == '__main__':
  main()
